#########################################
#File MA.R
#Creator: Matthew Wheeler
#Modified Date: 9/4/2017
##########################################


###############################################################
#Fit Individual Exponentially Modified Normal Model
#Inputs:
# data   - survival data
# frange - range of the survival function interested in
#Output:
# fit  - Fitted survival model
# mean - mean estimated survival curve - integrated over the random effects
# ub   - 100*(1-alpha)% upper bound    - integrated over the random effects
# lb   - 100*(alpha)%   lower bound    - integrated over the random effects
###############################################################
gpareto_model_Analyze <- function(data,frange,alpha = 0.05, sample = 30000,mcmc_warmup=1000,adapt_delta=0.9,seed=8675309){
  t <- seq(0,1,0.0025)
  ####################################################
  #fit the actual model
  ###################################################
  fit <- gPareto_model_fit(data,mcmc_iter = sample,mcmc_warmup = mcmc_warmup,adapt_delta = adapt_delta,
                                      seed = seed)
  ##################################################
  #extract the parameters
  ##################################################
  est.parms <- extract(fit$modelFit)
  b <- est.parms$b
  l <- est.parms$l


  surv <- do.call(rbind,vgpareto_surv(t, l, b,length(l),length(t)))

  ###############################################################################
  #Create a survival function (mean), the upper bound of that survival function (ub)
  #and the lower bound of that survival functio (lb), this is done over the range of t
  ###############################################################################
  ub <- splinefun(t*frange[2],apply(surv,2,quantile,alpha,na.rm = T))
  mean <- splinefun(t*frange[2],apply(surv,2,mean,na.rm = T))
  lb <- splinefun(t*frange[2],apply(surv,2,quantile,1-alpha,na.rm = T))
  ##############################################################################
  return(list(fit=fit ,ub = ub,mean=mean,lb=lb))

}


###############################################################
#Fit Individual Exponentially Modified Normal Model
#Inputs:
# data   - survival data
# frange - range of the survival function interested in
#Output:
# fit  - Fitted survival model
# mean - mean estimated survival curve - integrated over the random effects
# ub   - 100*(1-alpha)% upper bound    - integrated over the random effects
# lb   - 100*(alpha)%   lower bound    - integrated over the random effects
###############################################################
expmNorm_model_Analyze <- function(data,frange,alpha = 0.05, sample = 30000,mcmc_warmup=1000,adapt_delta=0.9,seed=8675309){
  t <- seq(0,1,0.0025)
  ####################################################
  #fit the actual model
  ###################################################
  fit <- logExpModNorm_model_fit(data,mcmc_iter = sample,mcmc_warmup = mcmc_warmup,adapt_delta = adapt_delta,
                                 seed = seed)
  ##################################################
  #extract the parameters
  ##################################################
  est.parms <- extract(fit$modelFit)
  b <- est.parms$b
  l <- est.parms$l
  nu <- est.parms$nu

  surv <- do.call(rbind,vexpmn_surv(t,nu, l, b,length(l),length(t)))

  ###############################################################################
  #Create a survival function (mean), the upper bound of that survival function (ub)
  #and the lower bound of that survival functio (lb), this is done over the range of t
  ###############################################################################
  ub <- splinefun(t*frange[2],apply(surv,2,quantile,alpha,na.rm = T))
  mean <- splinefun(t*frange[2],apply(surv,2,mean,na.rm = T))
  lb <- splinefun(t*frange[2],apply(surv,2,quantile,1-alpha,na.rm = T))
  ##############################################################################
  return(list(fit=fit ,ub = ub,mean=mean,lb=lb))

}

###############################################################
#Fit Individual double exponential  Model
#Inputs:
# data   - survival data
# frange - range of the survival function interested in
#Output:
# fit  - Fitted survival model
# mean - mean estimated survival curve - integrated over the random effects
# ub   - 100*(1-alpha)% upper bound    - integrated over the random effects
# lb   - 100*(alpha)%   lower bound    - integrated over the random effects
###############################################################
dexp_model_Analyze <- function(data,frange,alpha = 0.05, sample = 30000,mcmc_warmup=1000,adapt_delta=0.9,seed=8675309){

  t <- seq(0,1,0.0025)
  ####################################################
  #fit the actual model
  ###################################################
  fit  <- logDoubleExp_model_fit(data,mcmc_iter = sample,mcmc_warmup = mcmc_warmup,adapt_delta = adapt_delta,
                                 seed = seed)
  est.parms <- extract(fit$modelFit)

  b <- est.parms$b
  l <- est.parms$l
  surv <- do.call(rbind,vdexp_surv(t, l, b, length(l),length(t)))
  ###############################################################################
  #Create a survival function (mean), the upper bound of that survival function (ub)
  #and the lower bound of that survival functio (lb), this is done over the range of t
  ###############################################################################

  ub <- splinefun(t*frange[2],apply(surv,2,quantile,alpha,na.rm = T))
  mean <- splinefun(t*frange[2],apply(surv,2,mean,na.rm = T))
  lb <- splinefun(t*frange[2],apply(surv,2,quantile,1-alpha,na.rm = T))
  ##############################################################################
  return(list(fit=fit ,ub = ub,mean=mean,lb=lb))

}


###############################################################
#Fit Individual Weibull  Model
#Inputs:
# data   - survival data
# frange - range of the survival function interested in
#Output:
# fit  - Fitted survival model
# mean - mean estimated survival curve - integrated over the random effects
# ub   - 100*(1-alpha)% upper bound    - integrated over the random effects
# lb   - 100*(alpha)%   lower bound    - integrated over the random effects
###############################################################
weib_model_Analyze <- function(data,frange,alpha = 0.05, sample = 30000,mcmc_warmup=1000,adapt_delta=0.9,seed=8675309){

  t <- seq(0,1,0.0025)
  ####################################################
  #fit the actual model
  ###################################################
  fit  <- weib_model_fit(data,mcmc_iter = sample,mcmc_warmup = mcmc_warmup,adapt_delta = adapt_delta,
                         seed = seed)
  est.parms <- extract(fit$modelFit)

  b <- est.parms$b
  l <- est.parms$l
  surv <- do.call(rbind, vweib_surv(t, l, b,length(l),length(t)))
  ###############################################################################
  #Create a survival function (mean), the upper bound of that survival function (ub)
  #and the lower bound of that survival functio (lb), this is done over the range of t
  ###############################################################################

  ub <- splinefun(t*frange[2],apply(surv,2,quantile,alpha,na.rm = T))
  mean <- splinefun(t*frange[2],apply(surv,2,mean,na.rm = T))
  lb <- splinefun(t*frange[2],apply(surv,2,quantile,1-alpha,na.rm = T))
  ##############################################################################
  return(list(fit=fit ,ub = ub,mean=mean,lb=lb))

}


###############################################################
#Fit Individual Log-Gaussian
#Inputs:
# data   - survival data
# frange - range of the survival function interested in
#Output:
# fit  - Fitted survival model
# mean - mean estimated survival curve - integrated over the random effects
# ub   - 100*(1-alpha)% upper bound    - integrated over the random effects
# lb   - 100*(alpha)%   lower bound    - integrated over the random effects
###############################################################
logGaussian_model_Analyze <- function(data,frange,alpha = 0.05, sample = 30000,mcmc_warmup=1000,adapt_delta=0.9,seed=8675309){

  t <- seq(0,1,0.0025)
  ####################################################
  #fit the actual model
  ###################################################
  fit  <- lognorm_model_fit(data,mcmc_iter = sample,mcmc_warmup = mcmc_warmup,adapt_delta = adapt_delta,
                            seed = seed)
  est.parms <- extract(fit$modelFit)


  b <- est.parms$b
  l <- est.parms$l
  surv <- do.call(rbind, vlnorm_surv(t, l, sqrt(1/b),length(l),length(t)))
  ###############################################################################
  #Create a survival function (mean), the upper bound of that survival function (ub)
  #and the lower bound of that survival functio (lb), this is done over the range of t
  ###############################################################################
  ub <- splinefun(t*frange[2],apply(surv,2,quantile,alpha,na.rm = T))
  mean <- splinefun(t*frange[2],apply(surv,2,mean,na.rm = T))
  lb <- splinefun(t*frange[2],apply(surv,2,quantile,1-alpha,na.rm = T))
  ##############################################################################
  return(list(fit=fit ,ub = ub,mean=mean,lb=lb))

}

###############################################################
#Fit Individual Inv-Gaussian
# Inputs:
# data   - survival data
# frange - range of the survival function interested in
# Output:
# fit  - Fitted survival model
# mean - mean estimated survival curve - integrated over the random effects
# ub   - 100*(1-alpha)% upper bound    - integrated over the random effects
# lb   - 100*(alpha)%   lower bound    - integrated over the random effects
###############################################################
invGaussian_model_Analyze <- function(data,frange,alpha = 0.05, sample = 30000,mcmc_warmup=1000,adapt_delta=0.9,seed=8675309){

  t <- seq(0,1,0.0025)
  ####################################################
  #fit the actual model
  ###################################################
  fit  <- invGaussian_model_fit(data,mcmc_iter = sample,mcmc_warmup = mcmc_warmup,adapt_delta = adapt_delta,
                                seed = seed)
  est.parms <- extract(fit$modelFit)

  b <- est.parms$b
  l <- est.parms$l
  surv <- do.call(rbind,vigauss_surv(t, l, b,length(l),length(t)))
  ###############################################################################
  #Create a survival function (mean), the upper bound of that survival function (ub)
  #and the lower bound of that survival functio (lb), this is done over the range of t
  ###############################################################################

  ub <- splinefun(t*frange[2],apply(surv,2,quantile,alpha,na.rm = T))
  mean <- splinefun(t*frange[2],apply(surv,2,mean,na.rm = T))
  lb <- splinefun(t*frange[2],apply(surv,2,quantile,1-alpha,na.rm = T))
  ##############################################################################
  return(list(fit=fit ,ub = ub,mean=mean,lb=lb))

}

###############################################################
#Fit Individual logPareto
# Inputs:
# data   - survival data
# frange - range of the survival function interested in
# Output:
# fit  - Fitted survival model
# mean - mean estimated survival curve - integrated over the random effects
# ub   - 100*(1-alpha)% upper bound    - integrated over the random effects
# lb   - 100*(alpha)%   lower bound    - integrated over the random effects
###############################################################
gPareto_model_Analyze <- function(data,frange,alpha = 0.05, sample = 30000,mcmc_warmup=1000,adapt_delta=0.9,seed=8675309){

  t <- c(seq(0,0.1,0.00025),seq(0.1005,0.5,0.005),seq(0.6,1.0,1))
  ####################################################
  #fit the actual model
  ###################################################
  fit  <- gPareto_model_fit(data,mcmc_iter = sample,mcmc_warmup = mcmc_warmup,adapt_delta = adapt_delta,
                              seed = seed)
  est.parms <- extract(fit$modelFit)

  b <- est.parms$b
  l <- est.parms$l
  surv <- do.call(rbind,vgpareto_surv(t, l, b,length(l),length(t)))
  ###############################################################################
  #Create a survival function (mean), the upper bound of that survival function (ub)
  #and the lower bound of that survival functio (lb), this is done over the range of t
  ###############################################################################

  ub <- splinefun(t*frange[2],apply(surv,2,quantile,alpha,na.rm = T))
  mean <- splinefun(t*frange[2],apply(surv,2,mean,na.rm = T))
  lb <- splinefun(t*frange[2],apply(surv,2,quantile,1-alpha,na.rm = T))
  ##############################################################################
  return(list(fit=fit ,ub = ub,mean=mean,lb=lb))
}

###############################################################
#Fit Individual log-Gumbel
# Inputs:
# data   - survival data
# frange - range of the survival function interested in
# Output:
# fit  - Fitted survival model
# mean - mean estimated survival curve - integrated over the random effects
# ub   - 100*(1-alpha)% upper bound    - integrated over the random effects
# lb   - 100*(alpha)%   lower bound    - integrated over the random effects
###############################################################
logGumbel_model_Analyze <- function(data,frange,alpha = 0.05, sample = 30000,mcmc_warmup=1000,adapt_delta=0.9,seed=8675309){

  t <- seq(0,1,0.0025)
  ####################################################
  #fit the actual model
  ###################################################
  fit  <- logGumbel_model_fit(data,mcmc_iter = sample,mcmc_warmup = mcmc_warmup,adapt_delta = adapt_delta,
                              seed = seed)
  est.parms <- extract(fit$modelFit)

  b <- est.parms$b
  l <- est.parms$l
  surv <- do.call(rbind,vlgum_surv(t, l, b,length(l),length(t)))
  ###############################################################################
  #Create a survival function (mean), the upper bound of that survival function (ub)
  #and the lower bound of that survival functio (lb), this is done over the range of t
  ###############################################################################

  ub <- splinefun(t*frange[2],apply(surv,2,quantile,alpha,na.rm = T))
  mean <- splinefun(t*frange[2],apply(surv,2,mean,na.rm = T))
  lb <- splinefun(t*frange[2],apply(surv,2,quantile,1-alpha,na.rm = T))
  ##############################################################################
  return(list(fit=fit ,ub = ub,mean=mean,lb=lb))
}

###############################################################
#Fit Individual log-Logistic
# Inputs:
# data   - survival data
# frange - range of the survival function interested in
# Output:
# fit  - Fitted survival model
# mean - mean estimated survival curve - integrated over the random effects
# ub   - 100*(1-alpha)% upper bound    - integrated over the random effects
# lb   - 100*(alpha)%   lower bound    - integrated over the random effects
###############################################################
logLogistic_model_Analyze <- function(data,frange,alpha = 0.05, sample = 30000,mcmc_warmup=1000,adapt_delta=0.9,seed=8675309){

  t <- seq(0,1,0.0025)
  ####################################################
  #fit the actual model
  ###################################################
  fit  <- logLogistic_model_fit(data,mcmc_iter = sample,mcmc_warmup = mcmc_warmup,adapt_delta = adapt_delta,
                              seed = seed)
  est.parms <- extract(fit$modelFit)

  b <- est.parms$b
  l <- est.parms$l
  surv <- do.call(rbind,vlogistic_surv(t, l, b,length(l),length(t)))
  ###############################################################################
  #Create a survival function (mean), the upper bound of that survival function (ub)
  #and the lower bound of that survival functio (lb), this is done over the range of t
  ###############################################################################

  ub <- splinefun(t*frange[2],apply(surv,2,quantile,alpha,na.rm = T))
  mean <- splinefun(t*frange[2],apply(surv,2,mean,na.rm = T))
  lb <- splinefun(t*frange[2],apply(surv,2,quantile,1-alpha,na.rm = T))
  ##############################################################################
  return(list(fit=fit ,ub = ub,mean=mean,lb=lb))
}


##################################################################################
#Function: MA_fit
#
##################################################################################
MA_fit <- function(data,max_data_t,method="pseudobma",alpha = 0.05, sample = 30000,mcmc_warmup=2000,adapt_delta=0.9,seed=8675309){

    data$t <- data$t/max_data_t
    frange = c(0,max_data_t)


    fit.llogit   <- logLogistic_model_Analyze(data,frange,alpha=alpha,sample=sample,mcmc_warmup = mcmc_warmup,adapt_delta=adapt_delta,
                                            seed = seed)
    fit.lGauss   <- logGaussian_model_Analyze(data,frange,alpha=alpha,sample=sample,mcmc_warmup = mcmc_warmup,adapt_delta=adapt_delta,
                                            seed = seed)
    fit.weibul   <- weib_model_Analyze(data,frange,alpha=alpha,sample=sample,mcmc_warmup = mcmc_warmup,adapt_delta=adapt_delta,
                                            seed = seed)
    fit.dexp     <- dexp_model_Analyze(data,frange,alpha=alpha,sample=sample,mcmc_warmup = mcmc_warmup,adapt_delta=adapt_delta,
                                            seed = seed)
    fit.lgpareto  <- gpareto_model_Analyze(data,frange,alpha=alpha,sample=sample,mcmc_warmup = mcmc_warmup,adapt_delta=adapt_delta,
                                            seed = seed)



    log_lik_list <- list()
    log_lik_list[[1]] = extract(fit.llogit$fit$modelFit)[["log_lik"]]
    log_lik_list[[2]] = extract(fit.lGauss$fit$modelFit)[["log_lik"]]
    log_lik_list[[3]] = extract(fit.weibul$fit$modelFit)[["log_lik"]]
    log_lik_list[[4]] = extract(fit.dexp$fit$modelFit)[["log_lik"]]
    log_lik_list[[5]] = extract(fit.lgpareto$fit$modelFit)[["log_lik"]]

    temp = rowSums(is.infinite(log_lik_list[[1]]))+ rowSums(is.infinite(log_lik_list[[2]]))+
           rowSums(is.infinite(log_lik_list[[3]]))+ rowSums(is.infinite(log_lik_list[[4]])) +
           rowSums(is.infinite(log_lik_list[[5]]));

    # Remove infinite values
    #
    #
    log_lik_list[[1]] = log_lik_list[[1]][temp==0,]
    log_lik_list[[2]] = log_lik_list[[2]][temp==0,]
    log_lik_list[[3]] = log_lik_list[[3]][temp==0,]
    log_lik_list[[4]] = log_lik_list[[4]][temp==0,]
    log_lik_list[[5]] = log_lik_list[[5]][temp==0,]

    model_weights <- loo_model_weights(log_lik_list,method=method,
                                        method,optim_method = "BFGS",BB=TRUE)


    ##############################################################################
    #
    ##############################################################################
    # find the max and adjust for numerical reasons
    ##############################################################################
    names(model_weights) <- c("Log-Logistic","Log-Gaussian","Weibull","Log-DoubleExponential",
                                "Gen Pareto")

    ##############################################################################
    #compute the functions for the model average
    ##############################################################################
    return( MA = list(fit.llogit=fit.llogit,fit.lGauss = fit.lGauss,fit.Weibul=fit.weibul,
                      fit.dexp = fit.dexp,  fit.gPareto = fit.lgpareto,
                      posterior.probs = model_weights,max = max_data_t) )
}

##############################################################
# This function returns a function with the estimated mean model
# average survival curve
#
###########################################################
est.meanMA_surv <- function(MAfits){
names(MAfits$posterior.probs) <- NULL
  return(meanMA <- function(x){
                   return( as.numeric(MAfits$posterior.probs[1]*MAfits$fit.llogit$mean(x) + MAfits$posterior.probs[2]*MAfits$fit.lGauss$mean(x) +
                           MAfits$posterior.probs[3]*MAfits$fit.Weib$mean(x)   + MAfits$posterior.probs[4]*MAfits$fit.dexp$mean(x) +
                           MAfits$posterior.probs[5]*MAfits$fit.expmNorm$mean(x) + MAfits$posterior.probs[6]*MAfits$fit.lGumbel$mean(x) +
                           MAfits$posterior.probs[7]*MAfits$fit.iGauss$mean(x)))
              })
}


##############################################################
# This function returns a function with the estimated mean model
# average survival curve along with the (1-alpha)% and alpha
# upper and lower bounds respectively
###########################################################
est.MAsurv_functions <- function(MAfits,frange,maxR,Q=0.5,alpha = 0.05){

  t <- c(seq(0,0.1,5e-5),seq(0.1001,0.3,0.001),seq(0.31,1,0.01))
  ##############################################################
  #log-normal
  est.parms <- extract(MAfits$fit.lGauss$fit$modelFit)
  b <- est.parms$b
  l <- est.parms$l
  mu <- rnorm(length(l),l,sqrt(est.parms$lsig_sq))
  lnorm.surv  <- do.call(rbind, vlnorm_surv(t, l, sqrt(1/b),length(l),length(t)))
  scurves <- array(0,c(dim(lnorm.surv),5))
  scurves2 <- scurves
  scurves2[,,1]  <- do.call(rbind, vlnorm_surv(t, mu, sqrt(1/b),length(l),length(t)))
  scurves[,,1] <- lnorm.surv
  ###################################################################################
  #double exponential

  est.parms <- extract(MAfits$fit.dexp$fit$modelFit)
  b <- est.parms$b
  l <- est.parms$l
  mu <- rnorm(length(l),l,sqrt(est.parms$lsig_sq))
  scurves[,,2] <-  do.call(rbind,vdexp_surv(t, l, b,length(l),length(t)))
  scurves2[,,2] <- do.call(rbind,vdexp_surv(t, mu, b,length(l),length(t)))
  ###########################################################
  #Weibull
  est.parms <- extract(MAfits$fit.Weibul$fit$modelFit)
  b <- est.parms$b
  l <- est.parms$l
  mu <- rnorm(length(l),l,sqrt(est.parms$lsig_sq))
  mu <- apply(est.parms$l_reff,1,mean)
  scurves[,,3] <- do.call(rbind, vweib_surv(t, l, b,length(l),length(t)))
  scurves2[,,3] <- do.call(rbind,vweib_surv(t, mu, b,length(l),length(t)))

  ##############################################################
  #generalized Pareto
  est.parms <- extract(MAfits$fit.gPareto$fit$modelFit)
  b <- est.parms$b
  l <- est.parms$l
  reff <- rnorm(length(est.parms$lsig_sq),0,sqrt(est.parms$lsig_sq))
  mu <- rnorm(length(l),l,sqrt(est.parms$lsig_sq))
  scurves[,,4] <-  do.call(rbind,vgpareto_surv(t, l, b,length(l),length(t)))
  scurves2[,,4] <- do.call(rbind,vgpareto_surv(t, mu, b,length(l),length(t)))
  ##################################################################
  ## log-logistic
  est.parms <- extract(MAfits$fit.llogit$fit$modelFit)
  b <- est.parms$b
  l <- est.parms$l
  mu <- rnorm(length(l),l,sqrt(est.parms$lsig_sq))
  scurves[,,5] <- do.call(rbind,vlogistic_surv(t, l, b,length(l),length(t)))
  scurves2[,,5] <- do.call(rbind,vlogistic_surv(t, mu, b,length(l),length(t)))
  ################################################################

  probs <- output$posterior.probs[c(2,4,3,4,1)] #reorder the posterior probabilities

  msample <- function(X,probs){
    return(X[,sample(1:5,1,prob=probs)])
  }

  surv2 <- t(apply(scurves,1,msample,probs=probs))
  surv3 <- t(apply(scurves2,1,msample,probs=probs))
 # plot(t,apply(lnorm.surv,2,mean))

  tempmean <- splinefun(t*maxR,apply(surv2,2,mean,na.rm = T))
  lb <- splinefun(t*maxR,apply(surv2,2,quantile,1-alpha,na.rm = T))
  ub <- splinefun(t*maxR,apply(surv2,2,quantile,alpha,na.rm = T))


  qtile <- splinefun(t*maxR,apply(surv3,2,mean,na.rm = T))
  ubq <- splinefun(t*maxR,apply(surv3,2,quantile,alpha,na.rm = T))
  lbq <- splinefun(t*maxR,apply(surv3,2,quantile,1-alpha,na.rm = T))
  return(functions = list(mean=tempmean,ub=ub,lb=lb,
                          qtile = qtile,ubq = ubq,lbq = lbq))
}
